function [EEG] = fillGaps(EEG,sbeforeTMS,safterTMS,sbeforerec,safterrec,imin,memb,nnei)
% Fills the gap, implementing a weighted forward and backward prediction by
% finding the k nearest neighbors of specific target-points (edges of the
% gap) from the segments prior to (forward prediction) and after the gap
% (backward prediction) from a kd-tree structure.
% INPUT
% -EEG            : EEG struct
% -nbeforeTMS     : index corresponding to the beginning of the gap
% -nafterTMS      : index corresponding to the end of the gap
% -sbeforerec     : duration of the segment for the forward prediction
% -safterrec      : duration of the segment for the backward prediction
% -imin           : index corresponding to the TMS event
% -memb           : number of embending dimensions
% -nnei           : number of nearest neighbors to be detected
% OUTPUT
% -EEG            : gap filled EEG struct

Tmax = (safterTMS-sbeforeTMS)*EEG.srate/1000+1; %gap duration
tau = 1; %latency step

fprintf('\n Calculating'); %annotate gap filling start
for b = 1 : size(EEG.data, 3)
    fprintf('.');
    xnewM = EEG.data(:,:,b)'; % transposed because of the knnsearch function
    for a = 1 : size(EEG.data, 1)
        
        x1V = xnewM(imin-sbeforerec:sbeforeTMS-1,a);
        x2V = xnewM(imin+safterrec:-1:safterTMS+1,a);
        
        % Forward prediction based on x1V
        nx1 = length(x1V);
        % State space reconstruction of the the training set
        nvec1 = nx1-1 - (memb-1)*tau;   % The length of the reconstructed set
        x1M = zeros(nvec1,memb);
        for iemb=1:memb
            x1M(:,memb-iemb+1) = x1V(1+(iemb-1)*tau:nvec1+(iemb-1)*tau);
        end
        y1V = x1V(2+(memb-1)*tau:nx1); % The one-step ahead mappings
        pre1V = NaN*ones(Tmax,1);
        winnowV = x1V(nx1-(memb-1)*tau:nx1);
        mdl = KDTreeSearcher(x1M); % Create the kd-tree
        for T = 1:Tmax
            % One step prediction
            tarV = winnowV((memb-1)*tau+1:-tau:1)'; %target-points array
            neiindV = knnsearch(mdl,tarV,'K',nnei); %find nearest neighbors
            yyV = y1V(neiindV);
            pre1V(T) = mean(yyV);
            winnowV = [winnowV(2:end);pre1V(T)];  % new one step window
        end
        
        % Backward prediction based on x1V
        nx2 = length(x2V);
        % State space reconstruction of the the training set
        nvec2 = nx2-1 - (memb-1)*tau;   % The length of the reconstructed set
        x2M = zeros(nvec2,memb);
        for iemb=1:memb
            x2M(:,memb-iemb+1) = x2V(1+(iemb-1)*tau:nvec2+(iemb-1)*tau);
        end
        y2V = x2V(2+(memb-1)*tau:nx2); % The one-step ahead mappings
        pre2V = NaN*ones(Tmax,1);
        winnowV = x2V(nx2-(memb-1)*tau:nx2);
        mdl2 = KDTreeSearcher(x2M); % Create kd-tree
        for T = 1:Tmax
            % One step prediction
            tarV2 = winnowV((memb-1)*tau+1:-tau:1)'; %target-points array
            neiindV = knnsearch(mdl2,tarV2,'K',nnei); %find nearest neighbors
            yyV = y2V(neiindV);
            pre2V(T) = mean(yyV); 
            winnowV = [winnowV(2:end);pre2V(T)]; % new one step window
        end
        
        % Give complementary weights according to their distance from start
        xnewM(sbeforeTMS:safterTMS,a)=((Tmax-1:-1:0)'.*pre1V+(0:Tmax-1)'.*flipud(pre2V))/(Tmax-1);
    end
    EEG.data(:,:,b) = xnewM';
end
    
end

